// 
// Module: AreaCalc
// Author: Hassan Siddiqui
//
// Modifications: See ChangeLog
// 
// $Id: AreaCalc.cc,v 1.18 2001/1/12 12:52:17 rsaxton Exp $

/* General classes */
#include <iostream>

/* SAS-specific classes */
#include "Tracer.h"
#include "NumericalRange.h"

/* Local Classes */
#include "AreaCalc.h"
#include "ArfDataServer.h"
#include "PixelIterator.h"

/* Constructor */
AreaCalc::AreaCalc()
{
  PROC;
  
      /* Get pointer to frame object */
      _frame  = arfDataServer->frame();

      /* Get pointer to detmap to be used in calculation. */
      _detmapPixelIterator  = arfDataServer->detmapPixelIterator();

      ENDPROC;
    };
  
/* Destructor */
AreaCalc::~AreaCalc()
{
  PROC; 
  ENDPROC;
};


/* Generate arf for a given input energy */
const vector<double> 
AreaCalc::area ( const vector<double>& energyVector ) 
{

  PROC;

 /* Return if there is a problem ith the detector map */
  if (!_detmapPixelIterator->hasValidNormalisation())
    {
      /* Return a zero area vector */
      errstr 
        << "The filtered detector map has a zero normalisation - "
        << "returning zero area. "
        << "Try using a flat detector map and/or more bins"
	<< Sas::warning("zeroSumDetmap");

      RETURN missingFluxCorr( vector<double>(energyVector.size(),0),
                arfDataServer->eeCorrector()->Fraction(energyVector), 
                arfDataServer->badPixCorrector()->TotalFactor(energyVector));

      //RETURN(vector<double>(energyVector.size(),0));
    }

 /* Get the out-of-time correction value */
  double ootFactor = arfDataServer->ootCorrection();

    errstr
        << "AreaCalc: ootFactor ="<<ootFactor
	<< Sas::message(Msg::App, Msg::Verbose);

 /* Define the quantum efficiency curve for later use */
  CcdQuantumEfficiencyCurve qecurve;
  bool qeset=false;

  /* The detector map can become unnormalised if one or more elements
     lies outside the CCDs used in the observation - redo the normalisation
     here */

  double renorm = 0;

  /* Get flux across detector */
  vector<double> fluxVector;
  
  for (unsigned int i = 0; i < energyVector.size(); ++i)
    {
      fluxVector.
	push_back(_frame->frameFlux(energyVector[i], 
				    arfDataServer->
				    framePeriod()));
    }
  
  vector<double> areaSumVector(energyVector.size());
  
  /* Loop over detmap pixels */

  FOREACH(_detmapPixelIterator)
    {
      PixelElement pixel = 
	_detmapPixelIterator->normalisedItem();

      errstr 
	<< "Computing contribution from pixel centred on DETX="
	<< pixel.xCentre()
	<< ", DETY=" 
	<< pixel.yCentre()
	<< ", value=" 
	<<  pixel.value()
	<< Sas::message(Msg::App, Msg::Verbose);

      
      /* If pixel has a value of zero, skip computation. */
      
      if ( pixel.value() >= 0 && pixel.value() <= FLT64_MIN )
	{
	  errstr 
	    << " AreaCalc: Pixel has zero value. Skipping computation " 
	    << Sas::message(Msg::App, Msg::Verbose);
	  
	  continue;
	}

      /* DETMAP defined in terms of CAMCOORD2 coordinates,
	 scaled to units of 0.05 arcsec. 
	 Need to convert this to CAMCOORD2 proper. */
      

      CamCoords camcoord2 = 
	arfDataServer->
	detxy005ToCamCoord2( pixel.xCentre(), 
			     pixel.yCentre());
      
      /* Find CCD corresponding to pixel. 
	 more on to next pixel if there is
	 no valid CCD found. */
      if (!arfDataServer->setCcd(camcoord2))
	{
          
          errstr<<"AreaCalc: Pixel off CCD. Skipping computation "
	    << Sas::message(Msg::App, Msg::Verbose);

	  continue;
	}

      /* Compute QE curve: The Q.E. curve is position dependent (at least
         for the PN and so we need to calculate the curve for each 
         grid point. Uncomment the qeset command below to save time by
         calculating Q.E. only once */
      if (! qeset) {


       /* Convert from CAMCOORD2 to PIXCOORD1 */
         PixCoords pixcoord = 
        	arfDataServer->camCoord2ToPixCoord(camcoord2);

         int raw_x = pixcoord.x();
         int raw_y = pixcoord.y();
         qecurve = arfDataServer->quantumEfficiencyCurve(raw_x, raw_y);

         // qeset = true;
      }

      /* Convert from CAMCOORD2 to TELCOORD */
      TelCoords telcoord = 
	arfDataServer->camCoord2ToTelCoord(camcoord2);
      double theta = telcoord.theta();
      double phi   = telcoord.phi();

      /* Compute effective area curve */
      EffectiveAreaCurve eacurve = 
	arfDataServer->effectiveAreaCurve(theta, phi);

      /* Compute filter transmission curve */
      FilterTransmissionCurve ftcurve =
	arfDataServer->filterTransmissionCurve(theta, phi);

      for (unsigned int i = 0; i < energyVector.size(); ++i)
	{
	  double weightedarea = 
	    pixelArea(energyVector[i],
                      pixel.value(),
		      fluxVector[i],
		      qecurve,
                      eacurve,
		      ftcurve,
                      ootFactor);

	  areaSumVector[i] += weightedarea;
	}

     /* Add this pixel to the renormalisation factor */
      renorm += pixel.value();

    }
      
  /* Now correct for missing flux due to encircled energy and bad pixels */

  areaSumVector = missingFluxCorr( areaSumVector, 
                arfDataServer->eeCorrector()->Fraction(energyVector), 
                arfDataServer->badPixCorrector()->TotalFactor(energyVector));
  
  /* Renormalise for detector map intensity */
  if (renorm == 0) {
     errstr<<"No flux was found within detector map"
           << Sas::warning("zeroRenorm");
  } 
  else 
  {
     for (unsigned int i = 0; i < energyVector.size(); ++i) {
          areaSumVector[i] = areaSumVector[i] / renorm;
     }
  }

  RETURN areaSumVector;
}

/*** Methods */

/* Compute area for given pixel position */
double AreaCalc::pixelArea( double energy, 
                            double detfrac,
			    double flux,
			    const CcdQuantumEfficiencyCurve & qecurve,	
                            const EffectiveAreaCurve & eacurve,
			    const FilterTransmissionCurve & ftcurve,
                            double ootFactor 
                            ) const
{
      PROC;

      /* Compute the photon rate per unit pixel
       * incident on the CCDs, and thence the effective
       * area per pixel. These are given by
       *
       * transmitted area per pixel a_trans = A_m * T * S * detfrac
       *
       * flux (per pixel per frame) f_p = a_trans * F_i
       *
       * and
       *
       * overall_area_per_pixel     = a_trans * Q * P(energy, f_p)
       *
       * where
       *
       * F_i       = flux incident on mirrors (photons/frame/cm2)
       * A_m       = mirror effective area (cm2)
       * T         = filter transmission (unitless)
       * S         = spatial density
       * Q         = quantum efficiency
       * P         = attenuation due to pileup (a function of f_p)
       * detfrac   = fraction of flux at this pixel position
       */

      double effarea =  eacurve.area( energy );

      double filtertrans = ftcurve.transmission( energy );

      double transmittedAreaPerPixel = 
	effarea * filtertrans * detfrac;
      
      double qe = qecurve.efficiency( energy );

      double pileupcorrn  = 
	arfDataServer->pileupCorrection 
	( energy,
	  transmittedAreaPerPixel * flux);
      
      double area= 
	transmittedAreaPerPixel       *
	qe * 
	pileupcorrn *
        ootFactor;

      RETURN area;
    }

/* Correct the area for encircled energy flux loss, taking into 
   account any bad pixels */

vector<double> AreaCalc::missingFluxCorr ( const vector<double> & area,
                         const vector<double> & eeCorrvector,
                         const vector<double> & badPixvector ) const
{
  PROC;

  vector<double> correctedarea;

  for (unsigned i = 0; i < eeCorrvector.size(); ++i)
    {

     /* Multiply the area, for this energy, by the PSF correction factor 
        any by the area loss factor */

      correctedarea.push_back( area[i] * 
                      ( eeCorrvector[i] - badPixvector[i] ) );
    }

  errstr<<"Bad pixel & chip gap fraction: ~"<<
              (badPixvector[0]/eeCorrvector[0]*100.0)<<" %"
	<< Sas::message(Msg::App, Msg::Sparse);
 
  RETURN correctedarea;
}


