//
// Module: BadPixelCorrector
// Author: Richard Saxton
//
// Description: 
//
// A class to provide corrections for bad pixels which are taken from
// extensions in the event file
//
// Modifications: See ChangeLog
// 
// $Id: BadPixelCorrector.cc,v 1.0 2001/1/11 15:12:49 rsaxton Exp $

/* General */

/* SAS-specific */
#include "Tracer.h"
#include "CalServer.h"
#include "Dal.h"
#include "OalServer.h"

/* Local */
#include "NumericalRange.h"
#include "BadPixelCorrector.h"
#include "BadPixPSFCorrector.h"
#include "PixelIterator.h"
#include "OffCCDPSFCorrector.h"

/* Constructor */
BadPixelCorrector::BadPixelCorrector(const string & badpixContainer, 
                   unsigned eefactor, PsfModel * psf)
  // :_eefactor(eefactor)
{
  PROC;
   
    /* Open the file containing the bad pixel values */

    if (dataSetServer->exists(badpixContainer)) {

      DataSet * bpset = dataSetServer->open( badpixContainer,
                                                 DataSet::Read );

      /* Set up an array of bad pixel correctors - one for each CCD */

      for (unsigned int ccd = 1; ccd <= 12; ccd++) {

        string bpstring = ccd < 10 ? "BADPIX0":"BADPIX";
        string table = bpstring+toString(ccd);

        errstr << "Checking for table: " << table
               << Sas::message(Msg::App, Msg::Verbose);

      /* If a bad pixel extension exists for this CCD, add it to the vector */
        if (bpset->hasBlock(table)) {
                         //fred = new BadPixPSFCorrector(table, eefactor, psf);
                         //_bpcorr->push_back(fred);

           _bpcorr.push_back(new BadPixPSFCorrector(badpixContainer+":"+table, 
                                               eefactor, psf, ccd));
  errstr<<"BadPix size now: "<<_bpcorr.size()
               << Sas::message(Msg::App, Msg::Verbose);
   
        }
      }
     /* Close the bad pixel set */
      dataSetServer->close(bpset);

    } 
    else { // No BadPixelFile specified or found

       errstr<<" Warning couldn't find dataset: "<<badpixContainer
             <<" - correcting for chip gaps but not for bad pixels"
             << Sas::warning("NoBadPixLocations");
    }

    /* Add an extra corrector for pixels which are outside the CCD boundaries */
     _bpcorr.push_back(new OffCCDPSFCorrector(eefactor, psf));

  errstr<<"BadPix size now: "<<_bpcorr.size()
               << Sas::message(Msg::App, Msg::Verbose);
   
  ENDPROC;
}

/* Destructor */
BadPixelCorrector::~BadPixelCorrector()
{
  PROC;
  ENDPROC;
}

/* Method: TotalFactor(energyVector)
         : Calculates total bad pixel effect folded through PSF etc. */

vector<double>
BadPixelCorrector::TotalFactor(const vector<double> & energyVector) const
{
  PROC;

   /* Define the corrections vector and initialise to zero */
    vector<double> corrFactor, tempvec;

    for (unsigned int i = 0; i < energyVector.size(); ++i) {
       corrFactor.push_back(0.0);
    }

  errstr<<"BadPix size: "<<_bpcorr.size()
               << Sas::message(Msg::App, Msg::Verbose);
   
   /* Filter the bad pixels to exclude those falling outside the 
      source region */
    for (unsigned int c= 0; c < _bpcorr.size(); c++) {
  errstr<<"BadPix filtering map: "<<c
               << Sas::message(Msg::App, Msg::Verbose);
   
        _bpcorr[c]->filterMap();
    }

   /* Add the correction factors together for each chip at each energy */
    for (unsigned int c= 0; c < _bpcorr.size(); c++) {
  errstr<<"BadPix calling Fraction: "<<c
               << Sas::message(Msg::App, Msg::Verbose);
   
         //corrFactor[i] += (*_bpcorr)[c]->Fraction(energyVector)[i];
        tempvec = _bpcorr[c]->Fraction(energyVector);

        for (unsigned int i = 0; i < energyVector.size(); ++i) {
            corrFactor[i] += tempvec[i];
        }
    }

  errstr<<"BadPix total 1:"<<corrFactor[0]<<" 5:"<<corrFactor[4]
               << Sas::message(Msg::App, Msg::Verbose);
   
  RETURN corrFactor;
}
  
int
BadPixelCorrector::totalPixels() const
{
  PROC;

    int total=0;

   /* Add up the discarded detector pixels from each element */ 
    for (unsigned int c= 0; c < _bpcorr.size(); c++) {

      errstr<<"BadPix - adding up pixels: "<<c
               << Sas::message(Msg::App, Msg::Verbose);
   
      total += _bpcorr[c]->totalPixels();
    
errstr<<"BPtot= "<<total
    << Sas::message(Msg::App, Msg::Verbose); 
    }
  
  RETURN total;
}
