//
// Module: CalPsfModel
// Author: Hassan Siddiqui
//
// Description: 
//
// Wrapper class around CAL psf model.
//
// Modifications: See ChangeLog
// 
// $Id: CalPsfModel.cc,v 1.4 2000/11/19 15:16:54 hsiddiqu Exp $

/* SAS */
#include "errstr.h"
#include "Tracer.h"
#include "CalServer.h"
#include "Region.h"

// Local classes
#include "ArfDataServer.h"
#include "CalPsfModel.h"
#include "PositionUtils.h"

/* Constructor */
CalPsfModel::CalPsfModel( double energy,
			  double detxcentre,
			  double detycentre,
                          AccuracyLevel mode,
			  const PositionUtils * const putils):
  _putils(putils),
  _mode(mode),
  _offset(0,0,0),
  _t(0,0),
  _boxShape(ShapeBasics::CIRCLE)
{
  PROC;

  errstr 
    << "Constructing CalPsfModel with energy=" << energy
    << " detxcentre=" << detxcentre
    << " detycentre=" << detycentre
    << " mode=" << _mode
    << Sas::message(Msg::App, Msg::Verbose);


  _offset = _putils->detxy005ToCamCoord2( detxcentre, 
					  detycentre );

  _t = _putils->camCoord2ToTelCoord(_offset);
  
  calServer.state()->accuracyLevel()->set(_mode);

  PsfDataServer * psfs = 0; psfs = calServer.getAtom(psfs);

  /* Use a nasty trick to convert a reference to a pointer. */
  errstr 
    << "Calling PsfDataServer for a Psf at TELCOORD position ["
    << setprecision(10) << _t.theta()
    << ","
    << setprecision(10) << _t.phi()
    << "]"
    << Sas::message(Msg::App, Msg::Verbose);
  
    Psf & p = psfs->psf(energy, _t.theta(), _t.phi());
  
    _psf = &p;

    /* Reduce bin size */
    if (_psf->binSizeChangeable())
      {

	const unsigned OVERSAMPLE_FACTOR = 4;

      	Psf::BinSize bz = _psf->binSize();
	Psf::BinSize bz_out(bz.binX()/OVERSAMPLE_FACTOR, bz.binY()/OVERSAMPLE_FACTOR);

	errstr 
	  << " Changing psf binsize from ["
	  << bz.binX() << "," << bz.binY()
	  << "] to [" 
	  << bz_out.binX() << "," << bz_out.binY() 
	  << "]"
	  << Sas::message(Msg::App, Msg::Noisy);
	_psf->binSize(bz_out);
      }

 
  ENDPROC;
};

CalPsfModel::~CalPsfModel()
{
  PROC;
  errstr 
    << "CalPsfModel destroyed."
    << Sas::message(Msg::App, Msg::Verbose);

  ENDPROC;
}

/* Return PSF function at given position */
double CalPsfModel::psf( double detx, 
			 double dety) const
{
  PROC;
  CamCoords c = _putils->detxy005ToCamCoord2( detx, 
					      dety );

  double out = (*_psf)(c.x(), c.y());

  errstr 
    << "CalPsfModel::psf: for detx=" << detx
    << " dety=" << dety
    << " get a psf of " << out
    << Sas::message(Msg::App, Msg::Noisy);

  RETURN out;
}

/* Calculate integrated PSF over a rectangular area. */
double CalPsfModel::int_psf( double detxboxcentre, 
			     double detyboxcentre, 
			     double detxboxsize, 
			     double detyboxsize ) const
{
  PROC;

  CamCoords c1 = _putils->detxy005ToCamCoord2( detxboxcentre, 
					       detyboxcentre );

  errstr 
    << "CalPsfModel::int_psf: setting box origin to ["
    << setprecision(10) << c1.x()-_offset.x()
    << ","
     << setprecision(10) << c1.y()-_offset.y()
    << "]"
    << Sas::message(Msg::App, Msg::Noisy);

  _psf->atOrigin(c1.x()-_offset.x(),c1.y()-_offset.y());

  /* Know that camcoords2/det have same origin (0,0).
     So can use same method for scales as well as positions */
  CamCoords c2 = _putils->detxy005ToCamCoord2( detxboxsize, 
					       detyboxsize );

  errstr 
    << "CalPsfModel::int_psf: creating a PsfWindow of "
    << setprecision(10) << c2.x()/2 << ","  
    << setprecision(10) << c2.y()/2
    << Sas::message(Msg::App, Msg::Noisy);

  Psf::PsfWindow pw(c2.x()/2., c2.y()/2);

  double out=0.0;

  
     errstr << "Mode"<<calServer.state()->accuracyLevel()->get()
     << Sas::message(Msg::App, Msg::Noisy);

  if (calServer.state()->accuracyLevel()->get() != EXTENDED) {

     out = _psf->enboxedEnergy(pw);

     errstr 
     << "CalPsfModel::int_psf: for "
     << " energy=" << _psf->E()
     << " theta=" << _t.theta()
     << " phi=" << _t.phi()
     << endl
     << " detxboxcentre=" << detxboxcentre
     << " detyboxcentre=" << detyboxcentre
     << " detxboxsize=" << detxboxsize
     << " detyboxsize=" << detyboxsize
     << endl
     << " derive a PsfWindow of "
     << " halfX=" << pw.halfX() 
     << " halfY=" << pw.halfY()
     << endl
     << " and get an enclosed energy " << out
     << Sas::message(Msg::App, Msg::Noisy);
  
   } else {   // circular or annular source box calculation

/* The above doesn't work very well for circular or annular source
   boxes - so use the following technique 
      Algorithm: Find the total PSF within the outer radius of
      the box, subtract the total PSF from within the inner radius
      and normalise for the box size over the annulus area
*/

   /* Calculate the mean radius of this box */
      double boxRadius = sqrt( (c1.x()-_offset.x()) * (c1.x()-_offset.x())
                        + (c1.y()-_offset.y()) * (c1.y()-_offset.y()) );

   /* This technique doesn't work for pixels right on the source
      centre so make sure radius is bigger than half a box */

      CamCoords bit = _putils->detxy005ToCamCoord2(1,1);

      if (boxRadius < c2.x()/2.0 + bit.x()) {
           boxRadius=c2.x()/2.0 + bit.x();
      }

   /* Find the PSF contribution in an annulus about this radius (in arcsec)  */
      double PsfCont = _psf->encircledEnergy(boxRadius + c2.x()/2.0) -
                       _psf->encircledEnergy(boxRadius - c2.x()/2.0); 

   /* Normalise by area to get the PSF fraction in the box */
      out = PsfCont * c2.x() * c2.y() / ( M_PI *
        ((boxRadius + c2.y()/2.0)*(boxRadius + c2.y() /2.0) -
         (boxRadius - c2.y()/2.0)*(boxRadius - c2.y()/2.0)) );

      errstr<<"Int_psf: " 
            <<"Centre: "<<detxboxcentre<<" "<<detyboxcentre
            <<" Widths: "<<detxboxsize<<" "<<detyboxsize
            <<"Radius: "<<boxRadius
            <<"psccont: "<<PsfCont
            <<"Normalised: "<<out
            << Sas::message(Msg::App, Msg::Noisy);
   }

  RETURN out;
}

/* Set energy */
void CalPsfModel::energy(double energy )
{
  PROC;

  _psf->setParameters(energy, _t.theta(), _t.phi());

  ENDPROC;
}

string CalPsfModel::mode() const
{
  PROC;

  RETURN "Dummy";
}

void CalPsfModel::setRadius( vector<double> radii )
{
  
  PROC;
    for (unsigned i=0; i<radii.size(); i++) {

      errstr
        << "Setting radius "<<i<<" = " <<radii[i]
        << Sas::message(Msg::App, Msg::Verbose);

       _radii.push_back(radii[i]);
    }
      errstr
        << "Finished for i="<<radii.size()
        << Sas::message(Msg::App, Msg::Verbose);
  ENDPROC;
}

void CalPsfModel::setNgridEls(unsigned int /* NGridEls */)
{
  PROC;
  /* Do nothing */
  ENDPROC;
}


double CalPsfModel::rad_psf() const
{
  PROC;

      double ee=0;

      errstr
        << "Getting Encircled energy (energy): "<<_psf->E()
        << "radius "<<_radii[0]
        << Sas::message(Msg::App, Msg::Verbose);

  /* Calculation method depends on source box shape */
      if ( _boxShape == ShapeBasics::CIRCLE ||
           _boxShape == ShapeBasics::ANNULUS) {

         ee = _psf->encircledEnergy(_radii[0]); 

  /* Subtract the encircled energy from within the inner radius if
     this is an annulus */

         if (_radii.size() == 2) {
            ee = _psf->encircledEnergy(_radii[1]) - ee;
         }

      } else {   // BOX

         double ee_inner = _psf->encircledEnergy(_radii[0]);

  /* Find the encircled energy for the circle touched by the corners of
     the rectangle */

         double diagonal = sqrt(_radii[0] * _radii[0] + _radii[1] * _radii[1]);

         double ee_outer = _psf->encircledEnergy(diagonal);

  /* Calculate the fraction of the annulus between these inner and outer
     circles which is contained within the rectangle */

         double afrac = ( 4.0 * _radii[0] * _radii[1] -
                         M_PI * _radii[0] * _radii[0] ) /
                     ( M_PI * diagonal * diagonal -
                         M_PI * _radii[0] * _radii[0] ); 

  /* Compute the total EEF */

         ee = ee_inner + (ee_outer - ee_inner) * afrac;

     }

  RETURN ee;
}

/* Method - setBoxShape() - sets the input shape */
void CalPsfModel::setBoxShape(ShapeBasics::ShapeType shape) 
{
  PROC;

     _boxShape = shape;

  ENDPROC;
}

